# -*- coding: utf-8 -*-
"""
Модуль обновления проектов и ПО через svn
(c)2013 ЗАО "РИПАС"
www.ripas.ru
"""
import svnupdater, pysvn, subprocess
CHECK_ERR = u'Ошибка при проверке обновления: '


class Updater:
    def __init__(self, local_path, new_rev_msg):
        if not hasattr(Updater, 'g_lock'):
            from threading import Lock
            Updater.g_lock = Lock()
        self.new_rev_msg = new_rev_msg
        from threading import Event
        self.request_update_event = Event()
        self.request_auth_event = Event()
        self.realm_event = Event()
        self.sleep_event = Event()
        self.update_period_event = Event()
        self.keepAuth = False
        self.worked = True
        self.lastUpdateError = None
        self.wcUpgraded = False
        self.username = None
        self.password = None
        del Event
        from threading import Thread
        self.update_thread = Thread(target=self.updateThread)
        del Thread
        from PythonQt import QtCore
        self.mtt = QtCore.QTimer()
        self.mtt.interval = 30000
        self.mtt.connect('timeout()', self.mainTreadTimer)
        del QtCore
        self.client = pysvn.Client()
        self.client.callback_get_login = self.get_login
        self.client.callback_ssl_server_trust_prompt = self.ssl_server_trust_prompt
        self.setLocalPath(local_path)

    def setLocalPath(self, local_path):
        self.local_path = local_path
        if len(local_path) == 0 and hasattr(self, 'tortoiseUpdateCmd'):
            del self.tortoiseUpdateCmd
        if svnupdater.TortoiseProc:
            self.tortoiseUpdateCmd = svnupdater.TortoiseProc + ' /command:update /path:"%s"' % self.local_path
        self.wcUpgraded = False
        if not self.update_period_event.isSet():
            self.update_period_event.set()

    def __del__(self):
        self.mtt.stop()
        self.worked = False
        if not self.update_period_event.isSet():
            self.update_period_event.set()
        if not self.sleep_event.isSet():
            self.sleep_event.set()
        if self.update_thread.is_alive():
            self.worked = False
            self.update_thread.join()

    def setAutoUpdate(self, start=True):
        if len(self.local_path) == 0:
            return
        if start:
            self.mtt.start()
            self.sleep_event.set()
            if not self.update_thread.is_alive():
                self.update_thread.start()
        else:
            self.sleep_event.clear()
            self.mtt.stop()

    def checkOnce(self):
        if len(self.local_path) == 0:
            return
        autoupdate_started = self.sleep_event.isSet()
        if autoupdate_started:
            self.setAutoUpdate(False)
        self.client.callback_get_login = self.get_login_sync
        try:
            update_res = self.checkForUpdate()
            if update_res == 1:
                self.mainTreadTimer()
            elif update_res == 0:
                from PythonQt import QtGui
                from gui import mainWindow
                QtGui.QMessageBox.information(mainWindow, u'Проверка обновлений', u'Текущая ревизия актуальна')
                del QtGui, mainWindow
            else:
                from PythonQt import QtGui
                from gui import mainWindow
                QtGui.QMessageBox.critical(mainWindow, u'Проверка обновлений', self.lastUpdateError)
                del QtGui, mainWindow
                self.lastUpdateError = None
        finally:
            self.client.callback_get_login = self.get_login

        if autoupdate_started:
            self.setAutoUpdate(True)

    def get_login(self, realm, username, may_save):
        self.realm = realm
        self.username = username
        self.may_save = may_save
        self.request_auth_event.set()
        self.realm_event.wait()
        if not self.username:
            return False, '', '', self.may_save
        psw = self.password
        self.password = None
        self.realm_event.clear()
        return True, self.username, psw, self.may_save

    def get_login_sync(self, realm, username, may_save):
        self.realm = realm
        self.username = username
        self.may_save = may_save
        self.requestAuthData()
        if not self.username:
            return False, '', '', self.may_save
        psw = self.password
        self.password = None
        return True, self.username, psw, self.may_save

    def ssl_server_trust_prompt(self, trust_dict):
        return True, trust_dict['failures'], True

    def checkForUpdate(self):
        if svnupdater.noPySvn: return 2
        while True:
            try:
                Updater.g_lock.acquire()
                local_info = self.client.info(self.local_path)
                remote_info = self.client.info2(local_info['url'], depth=svnupdater.pysvn.depth.empty)
                if remote_info[0][1]['rev'].number > local_info['revision'].number:
                    self.request_update_event.set()
                    return 1
                else:
                    return 0
            except:
                if not self.wcUpgraded:
                    self.client.upgrade(self.local_path)
                    self.wcUpgraded = True
                    continue
                import sys, traceback
                from PythonQt import QtGui
                e = sys.exc_info()
                self.lastUpdateError = CHECK_ERR + str(e[1])
            finally:
                Updater.g_lock.release()
            break
        return 2

    def updateThread(self):
        if svnupdater.noPySvn: return
        from time import sleep
        while self.worked:
            self.sleep_event.wait()
            if not self.request_update_event.isSet():
                self.checkForUpdate()
            self.update_period_event.wait(svnupdater.checkUpdatePeriodSec)
            self.update_period_event.clear()
        del sleep

    def defaultUpdateProc(self, repo_path):
        try:
            Updater.g_lock.acquire()
            if hasattr(self, 'tortoiseUpdateCmd'):
                subprocess.call(self.tortoiseUpdateCmd)
            else:
                self.client.update(repo_path)
        except:
            if self.update_thread.is_alive() and self.sleep_event.isSet():
                self.setAutoUpdate(False)
        finally:
            Updater.g_lock.release()

    def mainTreadTimer(self):
        if svnupdater.noPySvn: return

        if self.lastUpdateError:
            from gui import msgError
            msgError(self.lastUpdateError)
            self.lastUpdateError = None

        if self.update_thread.is_alive():
            self.update_thread.join(0.01)

        if self.request_update_event.isSet():
            from PythonQt import QtGui
            from gui import mainWindow
            auth_wnd_active = False
            for w in QtGui.QApplication.topLevelWidgets():
                if w.objectName=='auth_wnd':
                    if w.isVisible():
                        auth_wnd_active = True
                    break
            mb = None
            if not auth_wnd_active:
                mb = QtGui.QMessageBox(mainWindow)
                mb.setWindowTitle(u'Вопрос')
                mb.setText(self.new_rev_msg)
                mb.addButton(QtGui.QMessageBox.Yes)
                mb.addButton(QtGui.QMessageBox.No)
                dont_check_button = None
                if self.update_thread.is_alive() and self.sleep_event.isSet():
                    dont_check_button = mb.addButton(u'больше не проверять', QtGui.QMessageBox.RejectRole)
                result = getattr(mb, 'exec')()
                if result == QtGui.QMessageBox.Yes:
                    if hasattr(self, 'updateProc'):
                        self.updateProc(self, self.local_path)
                    else:
                        self.defaultUpdateProc(self.local_path)
                elif result == 0 and mb.clickedButton() == dont_check_button:
                    self.setAutoUpdate(False)
            self.request_update_event.clear()
            if mb is not None:
                mb.deleteLater()
                del mb
        elif self.request_auth_event.isSet():
            self.requestAuthData()

    def requestAuthData(self):
        try:
            import gui
            auth_wnd = gui.widgetFromUi('svnupdater/svnauthdlg.ui')
            del gui
            auth_wnd.findChild('QLabel', 'realm').text = self.realm
            auth_wnd.findChild('QLineEdit', 'user').text = self.username
            auth_wnd.findChild('QCheckBox', 'saveAuth').checked = self.may_save
            from PythonQt import QtGui
            if getattr(auth_wnd, 'exec')() == QtGui.QDialog.Accepted:
                self.username = auth_wnd.findChild('QLineEdit', 'user').text
                self.password = auth_wnd.findChild('QLineEdit', 'password').text
                self.may_save = auth_wnd.findChild('QCheckBox', 'saveAuth').checked
            else:
                if not self.keepAuth:
                    self.username = None
                    self.password = None
                if self.update_thread.is_alive():
                    stopAutoUpdate()
        except:
            if not self.keepAuth:
                self.username = None
                self.password = None
        self.realm_event.set()
        self.request_auth_event.clear()

class InstallerUpdater:
    def __init__(self, maintenance_prog):
        if not hasattr(Updater, 'g_lock'):
            from threading import Lock
            Updater.g_lock = Lock()
        self.maintenance_prog = maintenance_prog
        from threading import Event
        self.request_update_event = Event()
        self.sleep_event = Event()
        self.update_period_event = Event()
        self.worked = True
        self.lastUpdateError = None
        del Event
        from threading import Thread
        self.update_thread = Thread(target=self.updateThread)
        del Thread
        from PythonQt import QtCore
        self.mtt = QtCore.QTimer()
        self.mtt.interval = 30000
        self.mtt.connect('timeout()', self.mainTreadTimer)
        del QtCore

    def __del__(self):
        self.mtt.stop()
        self.worked = False
        if not self.update_period_event.isSet():
            self.update_period_event.set()
        if not self.sleep_event.isSet():
            self.sleep_event.set()
        if self.update_thread.is_alive():
            self.worked = False
            self.update_thread.join()

    def setAutoUpdate(self, start=True):
        if start:
            self.mtt.start()
            self.sleep_event.set()
            if not self.update_thread.is_alive():
                self.update_thread.start()
        else:
            self.sleep_event.clear()
            self.mtt.stop()

    def checkOnce(self):
        autoupdate_started = self.sleep_event.isSet()
        if autoupdate_started:
            self.setAutoUpdate(False)
        update_res = self.checkForUpdate()
        if update_res == 1:
            self.mainTreadTimer()
        elif update_res == 0:
            from PythonQt import QtGui
            from gui import mainWindow
            QtGui.QMessageBox.information(mainWindow, u'Проверка обновлений', u'Текущая ревизия актуальна')
            del QtGui, mainWindow
        else:
            from PythonQt import QtGui
            from gui import mainWindow
            QtGui.QMessageBox.critical(mainWindow, u'Проверка обновлений', self.lastUpdateError)
            del QtGui, mainWindow
            self.lastUpdateError = None

        if autoupdate_started:
            self.setAutoUpdate(True)

    def checkForUpdate(self):
        while True:
            try:
                Updater.g_lock.acquire()
                proc = subprocess.Popen([self.maintenance_prog,"ch","-g","ifw.*=false"],stdout=subprocess.PIPE,stderr=subprocess.PIPE,creationflags=subprocess.CREATE_NO_WINDOW)
                try:
                    stdout, stderr = proc.communicate(timeout=10)
                except TimeoutExpired:
                    proc.kill()
                    stdout, stderr = proc.communicate()
                if proc.returncode != 0:
                    if stderr != b'':
                        self.lastUpdateError = CHECK_ERR + stderr.decode()
                    elif stdout != b'':
                        self.lastUpdateError = CHECK_ERR + stdout.decode()
                    else:
                        self.lastUpdateError = CHECK_ERR + u'процесс обновления завершился с кодом %d' % proc.returncode
                    return 2
                if stdout is None or stdout == b'':
                    return 0
                self.request_update_event.set()
                return 1
            except:
                import sys, traceback
                from PythonQt import QtGui
                e = sys.exc_info()
                self.lastUpdateError = CHECK_ERR + str(e[1])
            finally:
                Updater.g_lock.release()
            break
        return 2

    def updateThread(self):
        while self.worked:
            self.sleep_event.wait()
            if not self.request_update_event.isSet():
                self.checkForUpdate()
            self.update_period_event.wait(svnupdater.checkUpdatePeriodSec)
            self.update_period_event.clear()

    def defaultUpdateProc(self):
        try:
            Updater.g_lock.acquire()
            subprocess.Popen([self.maintenance_prog,"--su"],startupinfo=subprocess.STARTUPINFO(dwFlags=subprocess.DETACHED_PROCESS))
        except:
            if self.update_thread.is_alive() and self.sleep_event.isSet():
                self.setAutoUpdate(False)
        finally:
            Updater.g_lock.release()

    def mainTreadTimer(self):
        if self.lastUpdateError:
            from gui import msgError
            msgError(self.lastUpdateError)
            self.lastUpdateError = None

        if self.update_thread.is_alive():
            self.update_thread.join(0.01)

        if self.request_update_event.isSet():
            from PythonQt import QtGui
            from gui import mainWindow
            auth_wnd_active = False
            for w in QtGui.QApplication.topLevelWidgets():
                if w.objectName=='auth_wnd':
                    if w.isVisible():
                        auth_wnd_active = True
                    break
            mb = None
            if not auth_wnd_active:
                mb = QtGui.QMessageBox(mainWindow)
                mb.setWindowTitle(u'Вопрос')
                mb.setText(u'Есть новая версия программы или одного из её компонентов. Заустить программу обновления?')
                mb.addButton(QtGui.QMessageBox.Yes)
                mb.addButton(QtGui.QMessageBox.No)
                dont_check_button = None
                if self.update_thread.is_alive() and self.sleep_event.isSet():
                    dont_check_button = mb.addButton(u'больше не проверять', QtGui.QMessageBox.RejectRole)
                result = getattr(mb, 'exec')()
                if result == QtGui.QMessageBox.Yes:
                    self.defaultUpdateProc()
                elif result == 0 and mb.clickedButton() == dont_check_button:
                    self.setAutoUpdate(False)
            self.request_update_event.clear()
            if mb is not None:
                mb.deleteLater()
                del mb
